// Mini Game Template by g.c.
// win32: g++ game.cpp -Wall -pedantic -lSDL -lopengl32 -lglu32 -lSDL_image
// linux: g++ game.cpp -Wall -pedantic -lSDL -lGL -lGLU -lSDL_image

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <stdint.h>

#include <vector>

#include <GL/gl.h>
#include <GL/glu.h>

// Sometimes this needs to be changed to just <SDL.h> and <SDL_image.h>
// When in doubt, use sdl-config to find out
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>

#ifdef _WIN32
#  include <windows.h>
#endif

using namespace std;

// Remove SDL main hook on Windows
#ifdef _WIN32
#  undef main
#endif

// Macros.
#define UNUSED(a) ((void)(a))

// Globals.
static int screen_width;
static int screen_height;
bool keys[SDLK_LAST];
float ratio;

unsigned int current_level = 0;

bool game_complete = false;

unsigned int textures[32];
#define TEX_WALL     0
#define TEX_FLOOR    1
#define TEX_FLOOR_X  2
#define TEX_STONE    3
#define TEX_STONE_HI 4
#define TEX_PLAYER   5
#define TEX_SPLASH   6
#define TEX_SPLASH_END 7

int map[16][16];
// brak kafelka
#define MAP_NONE    -1
#define MAP_WALL    TEX_WALL
#define MAP_FLOOR   TEX_FLOOR
#define MAP_FLOOR_X TEX_FLOOR_X

struct player_st {
  int x, y;
} map_player;

struct stone_st {
  int x, y;
  bool hi; // true - stone is in the right place

  static struct stone_st mk(int x, int y) {
    stone_st s = { x, y, false };
    return s;
  }
};

vector<stone_st> map_stones;

bool LoadMap(const char *filename);


// Functions.
static bool InitSDL(bool fullscreen = false, int width = 640, int height = 480);
static bool InitOpenGL();

unsigned int ImgToTexture(const char *filename);
void DrawQuad(float x, float y, float w, float h);
void DrawQuadRGBA(float x, float y, float w, float h, float r, float g, float b, float a);
void DrawQuadTexture(float x, float y, float w, float h, unsigned int texture_id);

static bool Events();
static void Logic();
static void Scene();

// ------------------------------------------------------------------
// EVENTS
// ------------------------------------------------------------------

// Events
static bool
Events() {
  SDL_Event ev;
  memset(&ev, 0, sizeof(ev));

  while(SDL_PollEvent(&ev)) {
    switch(ev.type) { 
      case SDL_KEYUP:
	keys[ev.key.keysym.sym] = false;	
	break;
      
      case SDL_KEYDOWN:
        if(ev.key.keysym.sym == SDLK_ESCAPE)
          return false;
	keys[ev.key.keysym.sym] = true;
	break;	
	
      case SDL_QUIT:
        return false;      
    }
  }
  
  return true;
}


// ------------------------------------------------------------------
// LOGIC
// ------------------------------------------------------------------

int GetStoneByXY(int x, int y) {\
  size_t s = map_stones.size();
  for(size_t i = 0; i < s; i++) {
    if(map_stones[i].x == x &&
       map_stones[i].y == y)
      return i;
  }

  return -1;
}

static bool
LevelCompleteCond() {
  size_t s = map_stones.size();
  for(size_t i = 0; i < s; i++) {
    if(!map_stones[i].hi)
      return false;
  }

  return true; 
}

static bool
LoadNextLevel() {
  current_level++;

  char file_name[256] = {0};
  snprintf(file_name, sizeof(file_name) - 1, 
      "maps/%u.txt", current_level);

  map_stones.clear();

  return LoadMap(file_name);
}

// Logic
static void
Logic() {
  
  int p_dx = 0; // player delta x,y
  int p_dy = 0;

  if(keys[SDLK_UP])         { p_dy = -1; keys[SDLK_UP]    = false; }
  else if(keys[SDLK_DOWN])  { p_dy = 1;  keys[SDLK_DOWN]  = false; }
  else if(keys[SDLK_LEFT])  { p_dx = -1; keys[SDLK_LEFT]  = false; }
  else if(keys[SDLK_RIGHT]) { p_dx = 1;  keys[SDLK_RIGHT] = false; }

  int p_nx = map_player.x + p_dx; // player new x,y
  int p_ny = map_player.y + p_dy;

  bool p_move_ok = true;

  if(map[p_ny][p_nx] == MAP_WALL ||
     map[p_ny][p_nx] == MAP_NONE) {
    p_move_ok = false;
  } else {
    int s_id = GetStoneByXY(p_nx, p_ny);
    if(s_id != -1) {
      int s_nx = p_nx + p_dx; // stone new x,y
      int s_ny = p_ny + p_dy;

      bool s_move_ok = true;

      if(map[s_ny][s_nx] == MAP_WALL ||
         map[s_ny][s_nx] == MAP_NONE) {
        s_move_ok = false;
      } else if(GetStoneByXY(s_nx, s_ny) != -1) {
        s_move_ok = false;
      }

      if(!s_move_ok) {
        // Cannot move the stone.
        p_move_ok = false;
      } else {
        // OK, move the stone.
        map_stones[s_id].x = s_nx;
        map_stones[s_id].y = s_ny;

        map_stones[s_id].hi = (map[s_ny][s_nx] == MAP_FLOOR_X);

        if(map_stones[s_id].hi) {
          // Is level complete?
          if(LevelCompleteCond()) {
            if(!LoadNextLevel()) {
              game_complete = true;           
            } else {
              return;
            }
          }
        }
      }
    }
  }

  if(p_move_ok) {
    map_player.x = p_nx;
    map_player.y = p_ny;
  }
}

// ------------------------------------------------------------------
// SCENE
// ------------------------------------------------------------------

// Scene
static void
Scene() {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glTranslatef(0.0f, 0.0f, -15.0f);
  glScalef(1.0f, -1.0f, 1.0f);

  // Background.
  float k = 6.0f;
  DrawQuadTexture(0.0f, 0.0f, 4.0f * k, 3.0f * k, 
      game_complete ? textures[TEX_SPLASH_END] : textures[TEX_SPLASH]
      );

  if(game_complete) {
    return;
  }

  // Draw map.
  glTranslatef(-7.5f, -7.5f, 0.0f);

  for(int j = 0; j < 16; j++) {
    for(int i = 0; i < 16; i++) {
      if(map[j][i] != MAP_NONE)
        DrawQuadTexture(
            (float)i, float(j),
            1.0f, 1.0f,
            textures[map[j][i]]);
    }
  }

  // Turn on alpha
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);

  // Draw player.
  DrawQuadTexture(
      (float)map_player.x, (float)map_player.y,
      1.0f, 1.0f, 
      textures[TEX_PLAYER]);

  // Draw stones.
  size_t s = map_stones.size();
  for(size_t i = 0; i < s; i++)
    DrawQuadTexture(
        (float)map_stones[i].x, (float)map_stones[i].y,
        1.0f, 1.0f,
        map_stones[i].hi ?
          textures[TEX_STONE_HI] :
          textures[TEX_STONE]);

  // Turn off
  glDisable(GL_BLEND);


}

// ------------------------------------------------------------------
// Init & utils
// ------------------------------------------------------------------

bool
LoadMap(const char *filename) {
  // -1 == 0xffffffff -> 0xff
  memset(map, 0xff, sizeof(map));

  FILE *f = fopen(filename, "r");
  if(!f) {
    fprintf(stderr, "error: could not open map file \"%s\"\n",
        filename);
    return false;
  }

  for(int j = 0; j < 16; j++) {
    char line[64];
    fgets(line, sizeof(line), f);
    if(feof(f)) {
      fprintf(stderr, "error: unexpected end of file \"%s\"\n",
          filename);
      fclose(f);
      return false;
    }

    for(int i = 0; i < 16; i++) {
      switch(line[i]) {
        case ' ': map[j][i] = MAP_NONE; break;
        case '#': map[j][i] = MAP_WALL; break;
        case '.': map[j][i] = MAP_FLOOR; break;
        case 'x': map[j][i] = MAP_FLOOR_X; break;
        case 'p': map_player.x = i;
                  map_player.y = j;
                  map[j][i] = MAP_FLOOR;
                  break;
        case 'o': map_stones.push_back(stone_st::mk(i, j));
                  // map_stones.push_back((stone_st){i, j, false});
                  map[j][i] = MAP_FLOOR; break;
        default:
          fprintf(stderr, "error: unknown map character \"%c\" in \"%s\"\n",
              line[i], filename);
          fclose(f);
          return false;
      }
    }
  }
 
  printf("info: loaded map \"%s\"\n", filename);

  fclose(f);
  return true;
}


// main
int 
main(int argc, char **argv, char **envp) {
  // Unused.
  UNUSED(argc);
  UNUSED(argv);
  UNUSED(envp);

  // Init SDL.
  if(!InitSDL())    
    return 1;

  // Init OpenGL.
  if(!InitOpenGL()) 
    return 2;

  // Title
  SDL_WM_SetCaption("My game", "My game");

  // Textures.
  textures[TEX_WALL]    = ImgToTexture("gfx/murek.png");
  textures[TEX_FLOOR]   = ImgToTexture("gfx/kafelka.png");
  textures[TEX_FLOOR_X] = ImgToTexture("gfx/kafelka2.png");
  textures[TEX_STONE]   = ImgToTexture("gfx/kamien1.png");
  textures[TEX_STONE_HI]= ImgToTexture("gfx/kamien_hi1.png");
  textures[TEX_PLAYER]  = ImgToTexture("gfx/user.png");
  textures[TEX_SPLASH]  = ImgToTexture("gfx/splash.png");  
  textures[TEX_SPLASH_END]  = ImgToTexture("gfx/splash_end.png"); 

  LoadNextLevel();

  // Main loop:
  unsigned int last_time = SDL_GetTicks();
  ratio = 0.0f;

  for(;;) {

    // Main stuff.
    if(!Events())
      break;

    Logic();
    Scene();
    SDL_GL_SwapBuffers(); 

    // Calc time.
    unsigned int curr_time = SDL_GetTicks();
    ratio = (float)(curr_time - last_time) / 1000.0f;
    last_time = curr_time;
  }

  // Done.
  SDL_Quit();
  return 0;
}

// InitSDL
static bool
InitSDL(bool fullscreen, int width, int height) {
  int screen_bpp;
  int screen_flag = SDL_OPENGL;
  if(fullscreen) screen_flag |= SDL_FULLSCREEN;

  screen_width = width;
  screen_height = height;
 
  const SDL_VideoInfo *info;
  SDL_Surface *surface;  

  if(SDL_Init(SDL_INIT_VIDEO) < 0)
    return false;
 
  info = SDL_GetVideoInfo();
  if(info == NULL)
    goto err;

  screen_bpp = info->vfmt->BitsPerPixel;
  SDL_GL_SetAttribute(SDL_GL_RED_SIZE,   8);
  SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,  8);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

  SDL_putenv("SDL_VIDEO_CENTERED=center");

  surface = SDL_SetVideoMode(
            screen_width, screen_height, 
            screen_bpp, screen_flag);

  if(surface == NULL)
    goto err;

  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

  return true;
  
err:
  SDL_Quit();
  return false;
}

// InitOpenGL
static bool
InitOpenGL()
{
  float ratio;
  ratio = (float)screen_width / (float)screen_height;
  
  glShadeModel(GL_SMOOTH);  
  glClearColor(0, 0, 0, 0);  
  glViewport(0, 0, screen_width, screen_height);  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity( );
  gluPerspective( 60.0, ratio, 1.0, 1024.0 );
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glDisable(GL_DEPTH_TEST); // uncomment this if going 2D
  return true;
}


// Note: older version of opengl required image width and height
// to be a power of 2 (e.g. 16x32 or 512x128 or 256x256, etc).
unsigned int
ImgToTexture(const char *filename) {
  // Load image.
  SDL_Surface *img = IMG_Load(filename);
  if(!img) {
    fprintf(stderr, "error: ImgToTexture could not load image \"%s\"\n",
            filename);
    return ~0;
  }

  // Image type? Only 24 and 32 supported, rest must be converted.
  unsigned int img_type = 0;

  if(img->format->BitsPerPixel == 32) {
    img_type = GL_RGBA;
  } else if(img->format->BitsPerPixel == 24) {
    img_type = GL_RGB;
  } else {

    // Convert to 32 bits.
    SDL_PixelFormat fmt = {
      NULL, 32, 4, 
      0, 0, 0, 0,
      0, 8, 16, 24,
      0xff, 0xff00, 0xff0000, 0xff000000,
      0,
      0xff
    };

    SDL_Surface *nimg = SDL_ConvertSurface(img, &fmt, SDL_SWSURFACE);
    SDL_FreeSurface(img);

    if(!nimg) {
      fprintf(stderr, "error: ImgToTexture could not convert image \"%s\" to 32-bit\n",
              filename);
      return ~0;
    }

    // Done converting.
    img = nimg;
    img_type = GL_RGBA;
  }

  // Create texture.
  unsigned int texture_id = ~0;
  glGenTextures(1, &texture_id);
  glBindTexture(GL_TEXTURE_2D, texture_id);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  // You might want to play with these parameters.
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  
  // Gen texture.
  glTexImage2D(GL_TEXTURE_2D, 0, img_type, img->w, img->h,
    0, img_type, GL_UNSIGNED_BYTE, img->pixels);

  printf("info: ImgToTexture loaded texture \"%s\" as %u\n",
      filename, texture_id);

  // Done
  SDL_FreeSurface(img);
  return texture_id;
}

void
DrawQuad(float x, float y, float w, float h)
{
  w /= 2.0f; h /= 2.0f;

  glBegin(GL_QUADS);
  glTexCoord2f(0.0f, 0.0f); glVertex2f(x - w, y - h);
  glTexCoord2f(1.0f, 0.0f); glVertex2f(x + w, y - h);
  glTexCoord2f(1.0f, 1.0f); glVertex2f(x + w, y + h);
  glTexCoord2f(0.0f, 1.0f); glVertex2f(x - w, y + h);
  glEnd();
}

void
DrawQuadRGBA(float x, float y, float w, float h, float r, float g, float b, float a)
{
  // Get old color.
  float current_color[4];
  glGetFloatv(GL_CURRENT_COLOR, current_color);

  // Set new color and draw quad.
  glColor4f(r, g, b, a);
  DrawQuad(x, y, w, h);

  // Set old color.
  glColor4fv(current_color);
}

void
DrawQuadTexture(float x, float y, float w, float h, unsigned int texture_id)
{
  // Enable texturing if needed.
  bool texturing_enabled = glIsEnabled(GL_TEXTURE_2D);
  if(!texturing_enabled)
    glEnable(GL_TEXTURE_2D);

  // Bind texture and draw.
  glBindTexture(GL_TEXTURE_2D, texture_id);
  DrawQuad(x, y, w, h);

  // Disable if was disable.
  if(!texturing_enabled)
    glDisable(GL_TEXTURE_2D);
}

